/**
* Zeigt durch roboterCodingTask definierte Level zum Lösen an
*/
function TaskUI(
    nodeContainer       // DOM element to occupy
) {
    nodeContainer.innerHTML = '';
    this._msgPlaceGeneral = new ScoTools.MessagePlace(true, true);
    this._msgPlaceGeneral.appendAsChild(nodeContainer);
    this._msgPlace = new ScoTools.MessagePlace(false, true);
    this._msgPlace.appendAsChild(nodeContainer);

    this._showTask('Wird geladen…');
    this._contentContainer = ScoTools.createHtml('div', nodeContainer, { 'class': 'codeAndTable' });
    this._codeView = new ClickableSourceCode();
    this._codeView.onchange = (line => { this.check(line); return false; });
    this._table = new TableOfVars();
    this._table.onaddfailed = (msg => this.addMsg(msg, ScoTools.Message.TYPE_INFO));
    [this._codeView, this._table].forEach(({ node }) => ScoTools.createHtml('div', this._contentContainer).appendChild(node) );

    const resetBtn = ScoTools.createHtml('button', ScoTools.createHtml('p', this._contentContainer.lastChild, { 'class': 'resetBtnPlace' }));
    ScoTools.createHtml('img', resetBtn, { alt: "⎌", src: "share/Edit-undo.svg" });
    ScoTools.createHtml(false, resetBtn, null, ' Speichertabelle zurücksetzen');
    resetBtn.addEventListener('click', () => this.resetTable());

    this.startSeed = 0;
}

// Verwendete ScoTools.MessagePlace-Instanz für generelle Meldungen
TaskUI.prototype._msgPlaceGeneral = null;

// Verwendete ScoTools.MessagePlace-Instanz für aktuelle Meldung
TaskUI.prototype._msgPlace = null;

// Verwendete ClickableSourceCode-Instanz
TaskUI.prototype._codeView = null;

// Verwendete TableOfVars-Instanz
TaskUI.prototype._table = null;

// Container (DOM-Knoten) für Hauptinhalt
TaskUI.prototype._contentContainer = null;

// Angezeigter Java-Quellcode
Object.defineProperty(TaskUI.prototype, 'code', {
    enumerable: true,
    set(code) { this._codeView.code = code; },
    get() { return this._codeView.code; }
});

// Funktion, die beim Freischalten eines neuen Levels aufgerufen werden soll
TaskUI.prototype.onlevelunlock = function(unlockedLevelCount) { }

// Was ist gerade der aktive Schritt?
TaskUI.prototype._active = 0;
Object.defineProperty(TaskUI.prototype, 'active', {
    enumerable: true,
    get() { return this._active; },
    set(v) {
        if (!Number.isFinite(v) || ( v >= this.levels.length - 1)) {
            this.done = true;
            this._active = Infinity;
        } else
            this._active = Math.floor(Math.max(Math.min(this.levels.length - 1, v), 0));
        this._startLevel(this._active);
        if ((typeof this.onlevelunlock == 'function') || (this.onlevelunlock instanceof Function))
            this.onlevelunlock.call(this, this._active);
    }
});

// Wurden Aufgabe bereits abgeschlossen?
TaskUI.prototype._done = false;
Object.defineProperty(TaskUI.prototype, 'done', {
    enumerable: true,
    get() { return this._done; },
    set(v) {
        if ((v ? true : false) <= this._done)
            return;
        this._done = true;
        this.addGeneralMsg('Aufgabe erfolgreich abgeschlossen', ScoTools.Message.TYPE_SUCCESS);
    }
});

// Funktion, die beim Erhalt von Strafpunkten aufgerufen werden soll
TaskUI.prototype.onnewfine = function(totalFine, additionalFine) { }

// Gesammelte Anzahl an Fehlerpunkten
TaskUI.prototype._fineCount = 0;
Object.defineProperty(TaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;

        const oldFine = this._fineCount;
        this._fineCount = v;
        if ((typeof this.onnewfine == 'function') || (this.onnewfine instanceof Function))
            this.onnewfine.call(this, v, v - oldFine);
    }
});

// Generierte Leveldaten
TaskUI.prototype._levels = [];
Object.defineProperty(TaskUI.prototype, 'levels', {
    enumerable: true,
    get() { return this._levels; },
    set(levels) {
        if (!(levels instanceof Array))
            return;
        this._levels = levels;
        this.active = 0;
    }
});

// Startwert für Pseudozufallszahlengenerierung
TaskUI.prototype._startSeed = 0;
Object.defineProperty(TaskUI.prototype, 'startSeed', {
    enumerable: true,
    get() { return this._startSeed; },
    set(v) {
        if ((v == this._startSeed) || !Number.isFinite(v) || (this._startSeed === v))
            return;
        this._startSeed = v;
        this._importTask(true);
    }
});

/**
* Einen Zufallszahlengenerationsfunktion für Ganzzahlen erstellen
* @return Function      Zufallszahlengenerationsfunktion
*/
TaskUI.prototype._getRandomIntFkt = function(
    useSeed     // Pseudozufallszahlengenerierung (mit Startwert this.startSeed) statt Math.random verwenden?
) {
    if (!useSeed)
        return ((min, max) => (Math.floor(Math.random() * (max - min + 1)) + min));
    let seed = this.startSeed;
    return ((min, max) => { seed = (seed * 9301 + 49297) % 233280; return Math.floor(seed / 233280 * (max - min + 1)) + min; });
}

/**
* Daten von window.TASK übernehmen
* @return undefined
*/
TaskUI.prototype._importTask = function(
    useSeed     // Pseudozufallszahlengenerierung (mit Startwert this.startSeed) statt Math.random verwenden?
) {
    if (!(window.TASK instanceof Object))
        return;
    let data = window.TASK;
    if (data instanceof Function)
        data = window.TASK.call(window.TASK, this._getRandomIntFkt(useSeed));
    if (!(data instanceof Object))
        return;
    if (data.notSelectable)
        this._codeView.notSelectable = data.notSelectable;
    if (data.code)
        this.code = data.code;
    ['classes', 'allowArrays', 'enforceScopes', 'enableOverload'].forEach(key => data[key] && (this._table[key] = data[key]));
    if (data.steps)
        this.levels = data.steps.slice();
    this._showTask(data.task ? data.task : 'Führe den Code aus, indem du immer die Auswirkungen der hervorgehobenen Zeile in der Speichertabelle einträgst und die nächste auszuführende Codezeile auswählst.');
}

/**
* Früher gespeicherte Werte wieder übernehmen
* @return undefined
*/
TaskUI.prototype.loadValues = function(
    isDone,                 // Wurde diese Aufgabe bereits abgeschlossen?
    activeStep,             // Anzahl der bereits abgeschlossen Schritte
    fineCount,              // Anzahl der gesammelten Fehlerpunkte
    seed                    // Startwert für die Pseudozufallszahlengenerierung
) {
    this.startSeed = seed;
    this._fineCount = (Number.isFinite(fineCount) ? fineCount : 0);
    this.active = isDone ? Infinity : activeStep;
}

/**
* Aufgabenstellung anzeigen
* @return undefined
*/
TaskUI.prototype._showTask = function(
    task        // Anzuzeigender HTML-Code
) {
    if (!task && this._nodeTask) {
        this._nodeTask.parentNode.removeChild(this._nodeTask);
        this._nodeTask = null;
        return;
    }
    if (this._nodeTask) {
        this._nodeTask.innerHTML = task;
        return
    }
    this._nodeTask = ScoTools.createHtml('div', null, { 'class': 'auftrag' });
    this._nodeTask.innerHTML = task;
    this._msgPlace.node.parentNode.insertBefore(this._nodeTask, this._msgPlaceGeneral.node.nextSibling);
}

/**
* Informationen über aktuelles Level ermitteln
* @return Object    Eintrag aus this.levels
*/
TaskUI.prototype.getLevelInfo = function() {
    const level = this.levels[this.active];
    return level ? level : null;
}

/**
* Die Speichertabelle zurücksetzen
* @return undefined
*/
TaskUI.prototype.resetTable = function() {
    const vars = this.levels.slice(0, this.active + 1).reduce((vars, level) => (((level instanceof Object) && level.vars) ? level.vars : vars), []);
    this._table.load(vars);
}

/**
* Level (neu) starten
* @return undefined
*/
TaskUI.prototype._startLevel = function(
    levelNr         // Nummer des zu startenden Levels (beginnend bei 0)
) {
    this.clearMsg();
    const level = this.getLevelInfo();
    if (level) {
        if (this._btnRestart && this._btnRestart.parentNode) {
            this._btnRestart.parentNode.replaceChild(this._contentContainer, this._btnRestart);
            this._btnRestart = null;
        }

        this._procLevelStart(level);
        if (level.task)
            this._showTask(level.task);
        if ((typeof level.start == 'function') || (level.start instanceof Function))
            level.start();
        if (level.line !== undefined)
            this._codeView.line = level.line;
        ['allowArrays', 'enforceScopes', 'enableOverload'].map(key => {
            if (level[key] !== undefined)
                this._table[key] = level[key];
        });
        if (level.vars)
            this._table.load(level.vars);
    } else {
        if (this._contentContainer.parentNode) {
            const btnRestart = ScoTools.createHtml('button', null, { 'class': 'fullwidth' }, 'Aufgabe erneut üben');
            btnRestart.addEventListener('click', () => {
                this._fineCount = 0;
                this._importTask();
            });
            this._contentContainer.parentNode.replaceChild(btnRestart, this._contentContainer);
            this._btnRestart = btnRestart;
        }
        this._showTask(this.createFinalReport());
    }
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
TaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Hervorragend! Du hast die Aufgabe fehlerfrei gelöst.';
    if (this.fineCount == 1)
        return 'Gut gemacht: Du hast die Aufgabe gelöst und nur <strong>einen Fehlerpunkt</strong> kassiert.';
    const arrZahlen = ['', '', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Du hast die Aufgabe gemeistert, dabei allerdings <strong>' + arrZahlen[this.fineCount] + ' Fehlerpunkte</strong> gesammelt.';
    return 'Du hast die Aufgabe gemeistert, dabei allerdings <strong>' + this.fineCount + ' Fehlerpunkte</strong> gesammelt.';
}

/**
* Dem Level ggf. mitteilen, dass es soeben (neu) gestartet wurde
* @return undefined
*/
TaskUI.prototype._procLevelStart = function(
    level       // Gestartetes Level
) {
    if ((level instanceof Object) && (level.onstart instanceof Function))
        level.onstart.call(level);
    if ((level instanceof Object) && (level.onstart instanceof Array))
        level.onstart.reduce((prev, checker) => (prev || ((checker instanceof Function) ? checker.call(level) : checker)), undefined);
}

/**
* Eine (weitere) Meldung dauerhaft anzeigen
* @return undefined
*/
TaskUI.prototype.addGeneralMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlaceGeneral.add(text, type);
}

/**
* Aktive Meldung löschen
* @return undefined
*/
TaskUI.prototype.clearMsg = function() {
    this._msgPlace.clear();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
TaskUI.prototype.addMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}

/**
* Bearbeitung des aktuellen Schritts korrigieren
* @return Boolean       Schritt richtig bearbeitet?
*/
TaskUI.prototype.check = function(
    nextLine        // Ausgewählte Zeilennummer
) {
    const level = this.levels[this.active + 1];
    if (!(level instanceof Object)) {
        ++this.active;
        return true;
    }
    if (Number.isFinite(level.line) && (level.line != nextLine) && ((level.altLines instanceof Array) ? (level.altLines.indexOf(nextLine) == -1) : true)) {
        this.addMsg('Das ist nicht die Zeile, in der es weitergeht…');
        ++this.fineCount;
        return false;
    }
    const diffs = level.vars ? this._table.compare(level.vars) : [];
    if (diffs && diffs.length) {
        //const lastDiff = diffs.pop();
        //this.addMsg('Fehler in Speichertabelle: ' + (diffs.length ? (diffs.join(', ') + ' und ') : '') + lastDiff);
        this.addMsg('Fehler in Speichertabelle: ' + diffs[0] + ((diffs.length > 1) ? ', …' : ''));
        ++this.fineCount;
        return false;
    }
    ++this.active;
    return true;
}
